# zyx =========================================================================

__all__ = [
    "db",
    "DB",
    "Document",
]

from ..core import _UtilLazyLoader 

class db(_UtilLazyLoader):
    pass
db.init("zyx.data", "db")

class DB(_UtilLazyLoader):
    pass
DB.init("zyx.data", "DB")

class Document(_UtilLazyLoader):
    pass
Document.init("zyx.data", "_Document")


# zyx ==============================================================================

__all__ = [
    "completion",
    "CompletionClient",
]

from typing import Literal, Optional, Union, Callable, List, Iterator
from zyx.core import BaseModel

CompletionTools = Literal["web", "calculator", "shell", "python"]
CompletionProviders = [
    "openai/",
    "anthropic/",
    "ollama/",
]

class CompletionClient:
    def __init__(self,
                 model : str = "openai/gpt-3.5-turbo",
                 response_model : Optional[BaseModel] = None,
                 
                 name : Optional[str] = None,
                 task : Optional[str] = None,
                 introduction : Optional[str] = None,
                 description : Optional[str] = None,
                 instructions : Optional[str] = None,
                 prevent_hallucinations : Optional[bool] = False,
                 prevent_prompt_injection : Optional[bool] = False,
                 
                 base_url : Optional[str] = None,
                 api_key : Optional[str] = None,
                 organization : Optional[str] = None,
                 max_tokens : Optional[int] = None,
                 temperature : Optional[float] = None,
                 
                 tools: Optional[Union[CompletionTools, Callable, List[Union[CompletionTools, Callable]]]] = None,
                 verbose : bool = False,
                 markdown : bool = True,
                 
                 *args,
                 **kwargs):
        
        self.verbose = verbose
        self.model = model
        self.base_url = base_url
        self.api_key = api_key
        self.response_model = response_model
        self.tools = tools
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.organization = organization
        self.name = name
        self.introduction = introduction
        self.description = description
        self.instructions = instructions
        self.task = task
        self.prevent_hallucinations = prevent_hallucinations
        self.prevent_prompt_injection = prevent_prompt_injection
        self.args = args
        self.kwargs = kwargs
        self.markdown = markdown
        self.agent_tools = None
        
        if self.verbose is True:
            from zyx.core import logger
            self.logger = logger
        if self.verbose is True:
            self.logger.info(f"Initializing completion client with model: {model}")
        if self.tools:
            if self.verbose is True:
                logger.info(f"Initializing completion client with tools: {tools}")
            self._build_tools()
        
        self._build_completion_agent()
        
    def _build_tools(self):
        if isinstance(self.tools, list):
            self.agent_tools = []
            for tool in self.tools:
                if callable(tool):
                    self.agent_tools.append(tool)
                elif tool == "web":
                    from phi.tools.duckduckgo import DuckDuckGo
                    from phi.tools.website import WebsiteTools
                    self.agent_tools.append(WebsiteTools())
                    self.agent_tools.append(DuckDuckGo())
                elif tool == "calculator":
                    from phi.tools.calculator import Calculator
                    self.agent_tools.append(Calculator())
                elif tool == "shell":
                    from phi.tools.shell import ShellTools
                    self.agent_tools.append(ShellTools(base_dir = "."))
                elif tool == "python":
                    from phi.tools.python import PythonTools
                    self.agent_tools.append(PythonTools(base_dir = "."))
        elif callable(self.tools):
            self.agent_tools = self.tools      
                         
    def _build_completion_agent(self):
        from phi.assistant.assistant import Assistant
        if self.base_url is not None:
            if self.verbose is True:
                self.logger.info("Attempting to initialize custom provider.")
            try:
                from phi.llm.openai.chat import OpenAIChat
                if self.agent_tools is not None:
                    self.assistant = Assistant(
                        llm = OpenAIChat(
                            model = self.model,
                            base_url = self.base_url,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            organization = self.organization,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        task = self.task,
                        tools = self.agent_tools,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown= self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs,
                    )
                else:
                    self.assistant = Assistant(
                        llm = OpenAIChat(
                            model = self.model,
                            base_url = self.base_url,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            organization = self.organization,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
            except Exception as e:
                if self.verbose is True:
                    self.logger.error(f"Failed to initialize custom provider. {e}")
                else:
                    raise e
        elif self.model.startswith("openai/"):
            self.model = self.model.split("/")[1]
            try:
                from phi.llm.openai.chat import OpenAIChat
                if self.agent_tools is not None:
                    self.assistant = Assistant(
                        llm = OpenAIChat(
                            model = self.model,
                            base_url = self.base_url,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            organization = self.organization,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        tools = self.agent_tools,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
                else:
                    self.assistant = Assistant(
                        llm = OpenAIChat(
                            model = self.model,
                            base_url = self.base_url,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            organization = self.organization,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
            except Exception as e:
                if self.verbose is True:
                    self.logger.error(f"Failed to initialize OpenAI provider. {e}")
                else:
                    raise e
        elif self.model.startswith("anthropic/"):
            self.model = self.model.split("/")[1]
            try:
                from phi.llm.anthropic.claude import Claude
                if self.agent_tools is not None:
                    self.assistant = Assistant(
                        llm = Claude(
                            model = self.model,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        tools = self.agent_tools,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
                else:
                    self.assistant = Assistant(
                        llm = Claude(
                            model = self.model,
                            api_key = self.api_key,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
            except Exception as e:
                if self.verbose is True:
                    self.logger.error(f"Failed to initialize Anthropic provider. {e}")
                else:
                    raise e
        elif self.model.startswith("ollama/"):
            self.model = self.model.split("/")[1]
            try:
                from phi.llm.ollama.openai import OllamaOpenAI
                if self.agent_tools is not None:
                    self.assistant = Assistant(
                        llm = OllamaOpenAI(
                            model = self.model,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        tools = self.agent_tools if self.agent_tools else None,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
                else:
                    self.assistant = Assistant(
                        llm = OllamaOpenAI(
                            model = self.model,
                            max_tokens = self.max_tokens,
                            temperature = self.temperature,
                            
                            show_tool_calls = self.verbose
                        ),
                        name = self.name,
                        introduction = self.introduction,
                        description = self.description,
                        instructions = self.instructions,
                        
                        task = self.task,
                        prevent_hallucinations = self.prevent_hallucinations,
                        prevent_prompt_injection= self.prevent_prompt_injection,
                        output_model = self.response_model,
                        markdown = self.markdown,
                        parse_output = False,
                        *self.args,
                        **self.kwargs           
                    )
            except Exception as e:
                if self.verbose is True:
                    self.logger.error(f"Failed to initialize Ollama provider. {e}")
                else:
                    raise e
        else:
            raise ValueError(f"Invalid model or provider, please use the litellm format ('openai/gpt-3.5-turbo'): {self.model}")
                
# --- completion -------------------------------------------------------------------

def completion(
    messages : Union[str, list[str]] = None,
    model : str = "openai/gpt-3.5-turbo",
    response_model : Optional[BaseModel] = None,
    
    name : Optional[str] = None,
    task : Optional[str] = None,
    introduction : Optional[str] = None,
    description : Optional[str] = None,
    instructions : Optional[str] = None,
    prevent_hallucinations : Optional[bool] = False,
    prevent_prompt_injection : Optional[bool] = False,
    
    base_url : Optional[str] = None,
    api_key : Optional[str] = None,
    organization : Optional[str] = None,
    max_tokens : Optional[int] = None,
    temperature : Optional[float] = None,
    
    tools: Optional[Union[CompletionTools, Callable, List[Union[CompletionTools, Callable]]]] = None,
    verbose : bool = False,
    markdown : bool = False,
    stream : bool = False,
    
    chat : Optional[bool] = False,
    
    *args,
    **kwargs
) -> Union[Iterator[str], str, BaseModel]:
    """A function built on top of PhiData's abstractions to provide a quick and unified chat completions
    client with both tool calling & structured output.
    
    Example:
        ```python
        import zyx
        
        zyx.completion("How are you today?")
        ```
        
    Args:
        messages (Union[str, list[str], optional): The message or messages to send to the completion model.
        model (str, optional): The completion model to use. Defaults to "openai/gpt-3.5-turbo".
        response_model (Optional[BaseModel], optional): The response model to use. Defaults to None.
        
        name (Optional[str], optional): The name of the completion agent. Defaults to None.
        task (Optional[str], optional): The task of the completion agent. Defaults to None.
        introduction (Optional[str], optional): The introduction of the completion agent. Defaults to None.
        description (Optional[str], optional): The description of the completion agent. Defaults to None.
        instructions (Optional[str], optional): The instructions of the completion agent. Defaults to None.
        system_prompt (Optional[str], optional): The system prompt of the completion agent. Defaults to None.
        prevent_hallucinations (Optional[bool], optional): Prevent hallucinations in the completion agent. Defaults to None.
        prevent_prompt_injection (Optional[bool], optional): Prevent prompt injection in the completion agent. Defaults to None.
        
        base_url (Optional[str], optional): The base url of the completion agent. Defaults to None.
        api_key (Optional[str], optional): The api key of the completion agent. Defaults to None.
        organization (Optional[str], optional): The organization of the completion agent. Defaults to None.
        max_tokens (Optional[int], optional): The max tokens of the completion agent. Defaults to None.
        temperature (Optional[float], optional): The temperature of the completion agent. Defaults to None.
        
        tools (Optional[Union[CompletionTools, Callable, List[Union[CompletionTools, Callable]]]], optional): The tools to use in the completion agent. Defaults to None.
        verbose (bool, optional): Whether to show verbose output. Defaults to False.
        markdown (bool, optional): Whether to use markdown in the completion agent. Defaults to True.
        stream (bool, optional): Whether to stream the completion agent. Defaults to False.
        
        chat (Optional[bool], optional): Start Chat in CLI Mode.
    """
    if messages is None:
        raise ValueError("Please provide a message or messages to send to the completion model.")
    if isinstance(messages, str):
        messages = [{"role" : "user", "content" : messages}]
    try:
        assistant_client = CompletionClient(
            model = model,
            response_model = response_model,
            
            name = name,
            task = task,
            introduction = introduction,
            description = description,
            instructions = instructions,
            prevent_hallucinations = prevent_hallucinations,
            prevent_prompt_injection = prevent_prompt_injection,
            
            base_url = base_url,
            api_key = api_key,
            organization = organization,
            max_tokens = max_tokens,
            temperature = temperature,
            
            tools = tools,
            verbose = verbose,
            markdown = markdown,
            stream = stream,
            
            *args,
            **kwargs
        )
    except Exception as e:
        raise e
    
    if chat is True:
        assistant_client.assistant.cli_app(markdown = markdown, exit_on = ["exit", "quit"])
    else:
        try:
            response = assistant_client.assistant.run(
                messages = messages,
                stream = stream,
            )
            return response
        except Exception as e:
            raise e
    
        
                    


# zyx ==============================================================================

__all__ = [
    "ai",
    "audio",
    "image",
    
    'logger',
    'BaseModel',
    'Field',
    "rich_console",
    
    "completion",
    "CompletionClient",
    "embeddings",
    "function",
    "instructor_completion",
    
    "db",
    "batch",
    "lightning",
    "reader",
    
    "paint",
]

import builtins
from rich import print
builtins.print = print

from .core import logger, BaseModel, Field, rich_console, batch, lightning, reader
from .ai import completion, CompletionClient, embeddings, function, instructor_completion
from .image import paint

from .data import db

from . import ai
from . import audio
from . import image


# zyx =========================================================================

from typing import List, Optional, Union, Literal, Callable
from sqlmodel import Field, SQLModel
import os

DBType = Literal["sql", "qdrant"]

# --- Database Models --------------------------------------------------------------

class Document(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    content: str
    filename: str
    hash: str

# --- Database Class --------------------------------------------------------------

class DB:
    """Base class for database operations."""
    def __init__(
        self, 
        db_type: DBType = "sql", 
        db_path: str = ":memory:",
        embedding_model: str = "text-embedding-ada-002",
        vector_size: int = 1536,
        qdrant_url: str = "http://localhost:6333",
        collection_name: str = "documents",
        chunk_size: int = 1000,
        token_counter: Optional[Callable[[str], int]] = None,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
    ):
        self.db_type = db_type
        self.db_path = db_path
        self.embedding_model = embedding_model
        self.vector_size = vector_size
        self.collection_name = collection_name
        self.chunk_size = chunk_size
        self.token_counter = token_counter
        self.api_key = api_key
        self.api_base = api_base

        if db_type == "sql":
            self._setup_sql_db()
        elif db_type == "qdrant":
            self._setup_qdrant_db(qdrant_url)

    def _setup_sql_db(self):
        from sqlalchemy.engine import create_engine
        self.engine = create_engine(f"sqlite:///{self.db_path}", echo=False)
        SQLModel.metadata.create_all(self.engine)

    def _setup_qdrant_db(self, qdrant_url: str):
        from qdrant_client.models import VectorParams, Distance
        from qdrant_client.qdrant_client import QdrantClient
        self.qdrant_client = QdrantClient(qdrant_url)
        self.qdrant_client.recreate_collection(
            collection_name=self.collection_name,
            vectors_config=VectorParams(size=self.vector_size, distance=Distance.COSINE),
        )

    def add(self, path: Union[str, List[str]]):
        """Add a document or a list of documents to the database.
        Utilizes chunking to split the document into smaller chunks.
        
        Args:
            path (Union[str, List[str]]): The path to the document file or a list of paths.
            
        Raises:
            ValueError: If the path is invalid.
        """
        if isinstance(path, str):
            paths = [path]
        else:
            paths = path

        for file_path in paths:
            contents = self._read_files(file_path)
            chunked_contents = self._chunk_contents(contents)
            if self.db_type == "sql":
                self._add_to_sql(chunked_contents)
            elif self.db_type == "qdrant":
                self._add_to_qdrant(chunked_contents)

    def _read_files(self, path: str) -> List[tuple]:
        if os.path.isfile(path):
            return [self._process_file(path)]
        elif os.path.isdir(path):
            contents = []
            for root, _, files in os.walk(path):
                for file_name in files:
                    file_path = os.path.join(root, file_name)
                    contents.append(self._process_file(file_path))
            return contents
        else:
            raise ValueError(f"Invalid path: {path}")

    def _process_file(self, file_path: str) -> tuple:
        import hashlib
        with open(file_path, 'r', encoding='utf-8') as file:
            content = file.read()
        filename = os.path.basename(file_path)
        file_hash = hashlib.md5(content.encode()).hexdigest()
        return (content, filename, file_hash)

    def _chunk_contents(self, contents: List[tuple]) -> List[tuple]:
        import hashlib
        from ..core.text import chunk
        chunked_contents = []
        for content, filename, file_hash in contents:
            chunks = chunk(content, self.chunk_size, self.token_counter)
            for i, chunk_content in enumerate(chunks[0]): 
                chunk_hash = hashlib.md5(chunk_content.encode()).hexdigest()
                chunked_contents.append((chunk_content, f"{filename}_chunk_{i}", chunk_hash))
        return chunked_contents

    def _add_to_sql(self, contents: List[tuple]):
        from sqlalchemy.orm import Session
        from sqlalchemy.sql.expression import select
        with Session(self.engine) as session:
            for content, filename, file_hash in contents:
                existing_doc = session.execute(select(Document).where(Document.hash == file_hash)).first()
                if not existing_doc:
                    document = Document(content=content, filename=filename, hash=file_hash)
                    session.add(document)
            session.commit()

    def _add_to_qdrant(self, contents: List[tuple]):
        from qdrant_client.models import PointStruct
        from ..ai import embeddings
        points = []
        contents_list = [content for content, _, _ in contents]
        embedding_response = embeddings(
            inputs=contents_list,
            model=self.embedding_model,
            api_key=self.api_key,
            api_base=self.api_base,
        )
        for i, ((content, filename, file_hash), embedding_data) in enumerate(zip(contents, embedding_response['data'])):
            embedding_vector = embedding_data['embedding']
            point = PointStruct(
                id=i,
                vector=embedding_vector,
                payload={"content": content, "filename": filename, "hash": file_hash}
            )
            points.append(point)
        
        self.qdrant_client.upsert(collection_name=self.collection_name, points=points)

    def search(self, query: str, top_k: int = 5) -> List[dict]:
        """Searches the database for documents similar to the query.
        
        Args:
            query (str): The query string.
            top_k (int, optional): The number of results to return. Defaults to 5.
            
        Returns:
            List[dict]: A list of dictionaries containing the content and filename.
        """
        if self.db_type == "sql":
            return self._search_sql(query, top_k)
        elif self.db_type == "qdrant":
            return self._search_qdrant(query, top_k)

    def _search_sql(self, query: str, top_k: int) -> List[dict]:
        from sqlalchemy.orm import Session
        from sqlalchemy.sql.expression import select
        with Session(self.engine) as session:
            statement = select(Document).where(Document.content.contains(query)).limit(top_k)
            results = session.execute(statement).all()
            return [{"content": doc.content, "filename": doc.filename} for doc in results]

    def _search_qdrant(self, query: str, top_k: int) -> List[dict]:
        from ..ai import embeddings
        embedding_response = embeddings(
            inputs=[query],
            model=self.embedding_model,
            api_key=self.api_key,
            api_base=self.api_base,
        )
        query_vector = embedding_response['data'][0]['embedding']
        search_result = self.qdrant_client.search(
            collection_name=self.collection_name,
            query_vector=query_vector,
            limit=top_k
        )
        return [{"content": hit.payload["content"], "filename": hit.payload["filename"]} for hit in search_result]

# ==============================================================================

def db(
    db_type: DBType = "sql",
    db_path: str = ":memory:",
    embedding_model: str = "text-embedding-ada-002",
    vector_size: int = 1536,
    qdrant_url: str = "http://localhost:6333",
    collection_name: str = "documents",
    chunk_size: int = 1000,
    token_counter: Optional[Callable[[str], int]] = None,
    api_key: Optional[str] = None,
    api_base: Optional[str] = None,
) -> DB:
    """
    Creates and returns a DB instance with the specified configuration.
    
    Example:
        ```python
        db = DB()
        
        # Add a document
        db.add("path/to/document.txt")
        
        # Search for similar documents
        results = db.search("query")
        ```

    Parameters:
        db_type (DBType): The type of database to use ("sql" or "qdrant").
        db_path (str): The path to the database file (for SQL) or ":memory:" for in-memory database.
        embedding_model (str): The name of the embedding model to use.
        vector_size (int): The size of the vector embeddings.
        qdrant_url (str): The URL of the Qdrant server (for Qdrant database).
        collection_name (str): The name of the collection in Qdrant.
        chunk_size (int): The maximum number of tokens per chunk.
        token_counter (Callable[[str], int], optional): A function to count tokens in a string.
        api_key (str, optional): API key for the embedding model.
        api_base (str, optional): API base URL for the embedding model.

    Returns:
        DB: An instance of the DB class.
    """
    return DB(
        db_type=db_type,
        db_path=db_path,
        embedding_model=embedding_model,
        vector_size=vector_size,
        qdrant_url=qdrant_url,
        collection_name=collection_name,
        chunk_size=chunk_size,
        token_counter=token_counter,
        api_key=api_key,
        api_base=api_base,
    )

# ==============================================================================



# zyx ==============================================================================

__all__ = [
    "cast",
    "classify",
    "completion",
    "CompletionClient",
    "embeddings",
    "extract",
    "function",
    "instructor_completion",
    "generate",
    "inference",
]

from ..core import _UtilLazyLoader

class completion(_UtilLazyLoader):
    pass
completion.init("zyx.ai._completion", "completion")

from ._completion import CompletionClient as CompletionClient

class embeddings(_UtilLazyLoader):
    pass
embeddings.init("zyx.ai._embeddings", "embeddings")

class cast(_UtilLazyLoader):
    pass
cast.init("marvin.ai.text", "cast")

class extract(_UtilLazyLoader):
    pass
extract.init("marvin.ai.text", "extract")

class classify(_UtilLazyLoader):
    pass
classify.init("marvin.ai.text", "classify")

class function(_UtilLazyLoader):
    pass
function.init("marvin.ai.text", "fn")

class generate(_UtilLazyLoader):
    pass
generate.init("marvin.ai.text", "generate")

class inference(_UtilLazyLoader):
    pass
inference.init("huggingface_hub.inference._client", "InferenceClient")

class instructor_completion(_UtilLazyLoader):
    pass
instructor_completion.init("zyx.ai._instructor_completion", "instructor_completion")


# zyx ==============================================================================

__all__ = [
    "cast",
    "classify",
    "completion",
    "CompletionClient",
    "embeddings",
    "extract",
    "function",
    "generate",
    "inference",
    "instructor_completion",
]

from ._completion import completion as completion
from ._completion import CompletionClient as CompletionClient
from ._embeddings import embeddings as embeddings
from ._instructor_completion import instructor_completion as instructor_completion

from huggingface_hub.inference._client import InferenceClient as inference

from marvin.ai.text import cast as cast
from marvin.ai.text import extract as extract
from marvin.ai.text import classify as classify
from marvin.ai.text import fn as function
from marvin.ai.text import generate as generate


# zyx ==============================================================================

__all__ = [
    "Audio",
    "speak",
    "transcribe",
]

from ..core import _UtilLazyLoader 

class Audio(_UtilLazyLoader):
    pass
Audio.init("marvin.types", "Audio")

class speak(_UtilLazyLoader):
    pass
speak.init("marvin.ai.audio", "speak")

class transcribe(_UtilLazyLoader):
    pass
transcribe.init("marvin.ai.audio", "transcribe")


# zyx ==============================================================================

__all__ = [
    "reader",
    "chunk"
]

from .decorators import batch
import os
from typing import Callable, List, Union

def read_text_file(file_path: str) -> str:
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()

def read_docx_file(file_path: str) -> str:
    import docx
    doc = docx.Document(file_path)
    text = ''
    for para in doc.paragraphs:
        text += para.text + '\n'
    return text

def read_pdf_file(file_path: str) -> str:
    import fitz
    doc = fitz.open(file_path)
    text = ''
    for page in doc:
        text += page.get_text()
    return text

def read_file(file_path: str) -> str:
    ext = os.path.splitext(file_path)[1].lower()
    if ext == '.txt':
        return read_text_file(file_path)
    elif ext == '.docx':
        return read_docx_file(file_path)
    elif ext == '.pdf':
        return read_pdf_file(file_path)
    else:
        try:
            return read_text_file(file_path)
        except:
            raise ValueError(f"Unsupported file format: {ext}")
        

@batch(batch_size=10, verbose=True, timeout=None)
async def read_files(file_paths: List[str]) -> List[str]:
    return [read_file(file_path) for file_path in file_paths]

def reader(path: str) -> Union[str, List[str]]:
    import asyncio
    if os.path.isfile(path):
        return read_file(path)
    elif os.path.isdir(path):
        file_paths = [os.path.join(path, file_name) for file_name in os.listdir(path)]
        return asyncio.run(read_files(file_paths))
    else:
        raise ValueError(f"Invalid path: {path}")
    
# ==============================================================================

# zyx ==============================================================================

async def chunk_text(
    text: Union[str, List[str]],
    chunk_size: int,
    token_counter: Callable[[str], int] = None,
    memoize: bool = True,
    verbose: bool = False
) -> Union[List[str], List[List[str]]]:
    """Split a text into semantically meaningful chunks of a specified size as determined by the provided token counter.

    Args:
        text (Union[str, List[str]]): The text to be chunked.
        chunk_size (int): The maximum number of tokens a chunk may contain.
        token_counter (Callable[[str], int]): A callable that takes a string and returns the number of tokens in it.
        memoize (bool, optional): Whether to memoize the token counter. Defaults to `True`.
        verbose (bool, optional): Whether to enable verbose mode. Defaults to `False`.

    Returns:
        Union[List[str], List[List[str]]]: A list of chunks up to `chunk_size`-tokens-long, with any whitespace used to split the text removed.
    """
    import semchunk
    import tiktoken
    if token_counter is None:
        encoding = tiktoken.encoding_for_model('gpt-4')
        token_counter = lambda x: len(encoding.encode(x))

    try:
        chunker = semchunk.chunkerify(token_counter, chunk_size, memoize=memoize)
    except ImportError:
        raise ImportError("The semchunk package is required for this function. Please install it via `pip install semchunk`.")

    async def process_texts(texts: List[str]) -> List[Union[List[str], List[List[str]]]]:
        return [chunker(t) for t in texts]
    
    if isinstance(text, str):
        text = [text]
    
    try:
        results = await process_texts(text)
    except Exception as e:
        raise e
    
    return results

def chunk(
    text: Union[str, List[str]],
    chunk_size: int,
    token_counter: Callable[[str], int] = None,
    memoize: bool = True,
    verbose: bool = False
) -> Union[List[str], List[List[str]]]:
    """Split a text into semantically meaningful chunks of a specified size as determined by the provided token counter.

    Args:
        text (Union[str, List[str]]): The text to be chunked.
        chunk_size (int): The maximum number of tokens a chunk may contain.
        token_counter (Callable[[str], int]): A callable that takes a string and returns the number of tokens in it.
        memoize (bool, optional): Whether to memoize the token counter. Defaults to `True`.
        verbose (bool, optional): Whether to enable verbose mode. Defaults to `False`.

    Returns:
        Union[List[str], List[List[str]]]: A list of chunks up to `chunk_size`-tokens-long, with any whitespace used to split the text removed.
    """
    import asyncio
    return asyncio.run(chunk_text(text, chunk_size, token_counter, memoize, verbose))

# ==================================================================================================



# zyx ==============================================================================# zyx ==============================================================================

__all__ = [
    'logger',
    'BaseModel',
    'Field',
    "rich_console",
    "_UtilLazyLoader"
    
    "batch",
    "lightning",
    "chunk",
    "reader",
]

from loguru import Logger as logger
from pydantic.main import BaseModel as BaseModel
from pydantic.fields import Field as Field
from rich.console import Console as rich_console
from ..core import _UtilLazyLoader as _UtilLazyLoader

from .decorators import batch as batch
from .decorators import lightning as lightning
from .text import chunk as chunk
from .text import reader as reader


# zyx ==============================================================================

__all__ = [
    "caption",
    "paint",
    "Image",
    "ImageUrl",
]

from ..core import _UtilLazyLoader

class caption(_UtilLazyLoader):
    pass
caption.init("marvin.ai.text", "caption")

class paint(_UtilLazyLoader):
    pass
paint.init("marvin.ai.images", "paint")

class Image(_UtilLazyLoader):
    pass
Image.init("marvin.types", "Image")

class ImageUrl(_UtilLazyLoader):
    pass
ImageUrl.init("marvin.types", "ImageUrl")



# zyx ==============================================================================

__all__ = [
    "embeddings"
]

from typing import List, Literal, Optional, Union

EmbeddingsProviders = ["google/", "openai/", "ollama/"]

def embeddings(
    inputs : Union[list[str], str],

    model : Optional[str] = "openai/text-embedding-ada-002",
    dimensions : Optional[int] = None,
    
    host : Optional[str] = None,
    api_key : Optional[str] = None,
    base_url : Optional[str] = None,
    organization : Optional[str] = None,
    encoding_format : Literal["float", "base64"] = "float",
    
    verbose : Optional[bool] = False,
    
    *args, **kwargs
):
    """Utilize an external service to retrieve embeddings for the inputs text.
    
    Args:
        inputs (Union[list[str], str]): The inputs text to embed.
        model (Optional[str], optional): The model to use for embeddings. Defaults to "openai/text-embedding-ada-002".
        dimensions (Optional[int], optional): The number of dimensions for the embeddings. Defaults to None.
        host (Optional[str], optional): The host for the Ollama service. Defaults to None.
        api_key (Optional[str], optional): The API key for the service. Defaults to None.
        base_url (Optional[str], optional): The base URL for the service. Defaults to None.
        organization (Optional[str], optional): The organization for the service. Defaults to None.
        encoding_format (Literal["float", "base64"], optional): The encoding format for the embeddings. Defaults to "float".
        verbose (Optional[bool], optional): Whether to log information. Defaults to False.
    """
    if not inputs:
        raise ValueError("Inputs text is required.")
    if verbose is True:
        from zyx.core import logger
        
    if not any([model.startswith(provider) for provider in EmbeddingsProviders]):
        from phi.embedder.openai import OpenAIEmbedder
        try:
            embedder = OpenAIEmbedder(
                dimensions = dimensions,
                model = model,
                encoding_format = encoding_format,
                api_key = api_key,
                base_url = base_url,
                organization = organization
            )
            
            if isinstance(inputs, List):
                embeddings = []
                for i in inputs:
                    result = embedder.get_embedding(i)
                    embeddings.append(result)
                return embeddings
            else:
                return embedder.get_embedding(inputs)
        except Exception as e:
            if verbose is True:
                logger.error(f"Failed to initialize OpenAI Embedder: {e}")
            raise e
    elif model.startswith("openai/"):
        from phi.embedder.openai import OpenAIEmbedder
        model = model[7:]
        try:
            embedder = OpenAIEmbedder(
                dimensions = dimensions if dimensions is not None else 1536,
                model = model,
                encoding_format = encoding_format,
                api_key = api_key,
                base_url = base_url,
                organization = organization
            )
            if isinstance(inputs, List):
                embeddings = []
                for i in inputs:
                    result = embedder.get_embedding(i)
                    embeddings.append(result)
                return embeddings
            else:
                return embedder.get_embedding(inputs)
        except Exception as e:
            if verbose is True:
                logger.error(f"Failed to initialize OpenAI Embedder: {e}")
            raise e
    elif model.startswith("google/"):
        model = model[8:]
        from phi.embedder.google import GeminiEmbedder
        try:
            embedder = GeminiEmbedder(
                dimensions = dimensions if dimensions is not None else 1536,
                model = model,
                api_key = api_key,
            )
            if isinstance(inputs, List):
                embeddings = []
                for i in inputs:
                    result = embedder.get_embedding(i)
                    embeddings.append(result)
                return embeddings
            else:
                return embedder.get_embedding(inputs)
        except Exception as e:
            if verbose is True:
                logger.error(f"Failed to initialize Google Embedder: {e}")
            raise e
    elif model.startswith("ollama/"):
        model = model[8:]
        from phi.embedder.ollama import OllamaEmbedder
        try:
            embedder = OllamaEmbedder(
                dimensions = dimensions if dimensions is not None else 4096,
                model = model,
                api_key = api_key,
                host = host
            )
            if isinstance(inputs, List):
                embeddings = []
                for i in inputs:
                    result = embedder.get_embedding(i)
                    embeddings.append(result)
                return embeddings
            else:
                return embedder.get_embedding(inputs)
        except Exception as e:
            if verbose is True:
                logger.error(f"Failed to initialize Ollama Embedder: {e}")
            raise e
        